// A generic multi-threaded raytracer.
// Written by Nils Liaaen Corneliusen 2016.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// https://www.ignorantus.com/

// gcc -O3 -ffast-math -Wall -std=gnu99 -o cputrace -lm cputrace.c

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <math.h>
#include <sys/time.h>
#include <pthread.h>
#include <unistd.h>

#include "bmp.h"
#include "vec.h"

#define OBJNUM      22
#define REFLECTIONS  3

#define FARDIST 256.0f

#define PI 3.1415926535f

typedef struct {
    vec3  obj_pos[OBJNUM];
    vec3  obj_col[OBJNUM];
    float obj_rad[OBJNUM];
    vec3  plane_norm;
    float plane_dist;
    vec3  plane_col;
    vec3  sky_col;
    vec3  eye;
    vec3  light_pos;
    vec3  light_col;
    vec3  startpos;
    float ax;
    float ay;
    float frame;
    float fx_offset;
    int   width;
    int   height;
} SceneInfo;

static vec3 trace_pixel( SceneInfo *si, int x, int y )
{
    vec2 v_texCoord = vec2_set( x/(float)si->width, y/(float)si->height );

    vec3 t = vec3_set( si->startpos.x + si->ax * v_texCoord.x,
                       si->startpos.y + si->ay * v_texCoord.y,
                       si->startpos.z );

    vec3 ray_dir = vec3_normalize( t );

    vec3 col = vec3_set( 0.0f, 0.0f, 0.0f );

    vec3 ray_pos = si->eye;

    for( int refl = 0; refl < REFLECTIONS; refl++ ) {

        float rt_dist = FARDIST;
        int rt_hit = -1;

        for( int i = 0; i < OBJNUM; i += 2 ) {

            vec3  v0 = vec3_sub( ray_pos, si->obj_pos[i+0] );
            vec3  v1 = vec3_sub( ray_pos, si->obj_pos[i+1] );
            float b0 = vec3_dot( v0, ray_dir );
            float b1 = vec3_dot( v1, ray_dir );
            float d0 = b0 * b0 - vec3_dot( v0, v0 ) + si->obj_rad[i+0];
            float d1 = b1 * b1 - vec3_dot( v1, v1 ) + si->obj_rad[i+1];
            float t0 = d0 < 0.0f ? 0.0f : -b0 - sqrt( d0 );
            float t1 = d1 < 0.0f ? 0.0f : -b1 - sqrt( d1 );

            bool r;
            r       = t0 >= rt_dist || t0 <= 0.0f;
            rt_dist = r ? rt_dist : t0;
            rt_hit  = r ? rt_hit  : i+0;
            r       = t1 >= rt_dist || t1 <= 0.0f;
            rt_dist = r ? rt_dist : t1;
            rt_hit  = r ? rt_hit  : i+1;

        }

        if( rt_hit == -1 && refl > 0 ) break;

        // plane with shadows
        if( rt_hit == -1 && refl == 0 ) {

            rt_dist = -(vec3_dot( ray_pos, si->plane_norm ) + si->plane_dist) / vec3_dot( ray_dir, si->plane_norm );

            if( rt_dist <= 0.0f ) {
                // plane miss, sky
                vec2 dxy = vec2_sub( v_texCoord, vec2_set( 0.5f, 0.5f ) );
                float dv = 1.0f - sqrt( vec2_dot( dxy, dxy ) );
                col = vec3_mul1( si->sky_col, dv );
                break;
            }

            //  plane hit
            ray_pos = vec3_add( ray_pos, vec3_mul1( ray_dir, rt_dist ) );
            vec3 l  = vec3_normalize( vec3_sub( si->light_pos, ray_pos ) );

            // check shadow
            int i;
            for( i = 0; i < OBJNUM; i += 2 ) {
                vec3  v0 = vec3_sub( ray_pos, si->obj_pos[i+0] );
                vec3  v1 = vec3_sub( ray_pos, si->obj_pos[i+1] );
                float b0 = vec3_dot( v0, l );
                float b1 = vec3_dot( v1, l );
                float d0 = b0 * b0 - vec3_dot( v0, v0 ) + si->obj_rad[i+0];
                float d1 = b1 * b1 - vec3_dot( v1, v1 ) + si->obj_rad[i+1];
                float t0 = -b0 - sqrtf( d0 );
                float t1 = -b1 - sqrtf( d1 );
                if( t0 > 0.0f || t1 > 0.0f ) break;
            }

            float diffuse = i == OBJNUM ? vec3_dot( si->plane_norm, l ) * 1.2f : 0.0f;
            col = vec3_mul1( si->plane_col, diffuse );
            break;

        }

        // sphere hit + reflection
        ray_pos  = vec3_add( ray_pos, vec3_mul1( ray_dir, rt_dist ) );
        vec3 n   = vec3_normalize( vec3_sub( ray_pos, si->obj_pos[rt_hit] ) );
        vec3 l   = vec3_normalize( vec3_sub( si->light_pos, ray_pos ) );

        float diffuse  = fmax( vec3_dot( n, l ), 0.0f );
        float specular = powf( fmax( vec3_dot( ray_dir, vec3_sub( l, vec3_mul1( vec3_mul1( n, 2.0f ), diffuse ) ) ), 0.0f ), 8.0f );

        col = vec3_add( col, vec3_add( vec3_mul1( si->light_col, specular ),
                                       vec3_mul1( si->obj_col[rt_hit], diffuse ) ) );

        ray_dir = vec3_reflect( ray_dir, n );

    }

    return col;
}

static void trace_lines( SceneInfo *si, bmp_argb *dstbmp, int startline, int endline )
{
    for( int y = startline; y <= endline; y++ ) {

        uint32_t *dstrow = (uint32_t *)(dstbmp->argb + y * dstbmp->stride);

        for( int x = 0; x < si->width; x++ ) {

            vec3 col = trace_pixel( si, x, y );

            col = vec3_mul1( vec3_fmin1( col, 1.0f ), 255.0f );
            uint32_t r = col.x;
            uint32_t g = col.y;
            uint32_t b = col.z;
            uint32_t rgba = (b<<16)|(g<<8)|r;
            *dstrow++ = rgba;

        }
    }
}


typedef struct {
    pthread_t tid;
    int id;
    SceneInfo *si;
    int startline, endline;
    bmp_argb *dstbmp;
} ThreadInfo;

pthread_barrier_t bar;

void *thread_func( void *data )
{
    ThreadInfo *ti = (ThreadInfo *)data;

    printf( "Thread %4d: %4d lines (%4d-%4d)\n", ti->id, ti->endline - ti->startline + 1, ti->startline, ti->endline );
        
    pthread_barrier_wait( &bar );

    // ...

    pthread_barrier_wait( &bar );

    trace_lines( ti->si, ti->dstbmp, ti->startline, ti->endline );

    return 0;
}


const vec3 cols[7] = {
    { 1.0f, 0.0f, 0.0f },
    { 1.0f, 1.0f, 0.0f },
    { 0.0f, 1.0f, 0.0f },
    { 0.0f, 1.0f, 1.0f },
    { 0.0f, 0.0f, 1.0f },
    { 1.0f, 0.0f, 1.0f },
    { 1.0f, 0.0f, 0.0f }
};

static vec3 h2rgb( float h )
{
    h = h - floor( h );
    float Slice     = 6.0f * h;
    float SliceInt  = floor( Slice );
    float SliceFrac = Slice - SliceInt;

    int i = (int)SliceInt;
    return vec3_mix( cols[i], cols[i+1], SliceFrac );
}


static uint64_t get_usec( void )
{
    struct timeval tv;
    gettimeofday( &tv, NULL );
    return ((uint64_t)tv.tv_sec*1000000 + (uint64_t)tv.tv_usec);
}


#define RAD_SUN    2.8f
#define RAD_PLANET 0.5f
#define DIST_INNER (RAD_SUN+1.5f*RAD_PLANET + 0.4f)

static vec3 rotate( float frac, float tilt, float radius )
{
    float yv    = sinf( tilt );
    float xv    = cosf( tilt );
    float sinxy = radius*sinf( frac * (PI*2.0f) );

    return vec3_set( sinxy * xv,
                     sinxy * yv,
                     -radius*cosf( frac * (PI*2.0f) ) );
}

#define DEFAULT_WIDTH   1920
#define DEFAULT_HEIGHT  1080
#define DEFAULT_THREADS    0

static void print_usage( const char *exec_name )
{
    printf( "Usage: %s [OPTIONS]\n\n", exec_name );
    printf( "Options:\n" );
    printf( "-o WxH      Output resolution, default %dx%d\n", DEFAULT_WIDTH, DEFAULT_HEIGHT );
    printf( "-t threads  Number of threads, default %d\n", DEFAULT_THREADS );
}


int main( int argc, char *argv[] )
{
    int threads = DEFAULT_THREADS;
    int width   = DEFAULT_WIDTH;
    int height  = DEFAULT_HEIGHT;

    int opt;
    while( (opt = getopt( argc, argv, "ho:t:" ) ) != -1 ) {
        switch( opt ) {
        default:
        case 'h':
            print_usage(argv[0]);
            return 0;
        case 'o': sscanf( optarg, "%dx%d", &width, &height ); break;
        case 't': threads = atoi( optarg );                   break;
        }
    }

    if( width <= 0 || height <= 0 ) {
        printf( "Illegal width/height!\n" );
        return 1;
    }

    if( threads < 0 || threads > height ) { 
        printf( "Illegal number of threads!\n" );
        return 1;
    }

    printf( "Threads: %d\n", threads );

    // Initialize scene
    SceneInfo si;

    si.width  = width;
    si.height = height;

    si.eye       = vec3_set(  0.0f,  0.0f, -8.0f );
    si.light_pos = vec3_set( -4.0f, -6.0f, -9.5f );
    si.light_col = vec3_set(  1.0f,  1.0f,  1.0f );

    si.ax  = 10.0f * width / height;
    si.ay  = 10.0f;

    si.startpos = vec3_set(  si.ax  * -0.5f - si.eye.x,
                             si.ay  * -0.5f - si.eye.y,
                            -si.eye.z );

    si.plane_norm = vec3_normalize( vec3_set( 0.0f, -1.0f, 0.0f ) );
    si.plane_dist = 6.5f;
    si.plane_col  = vec3_set( 0.0f, 1.0f, 0.0f );
    si.sky_col    = vec3_set( 0.0f, 0.0f, 1.0f );

    // Center sphere
    si.obj_pos[0] = vec3_set( 0.0f, 0.0f, 0.0f );
    si.obj_rad[0] = RAD_SUN * RAD_SUN;
    si.obj_col[0] = vec3_set( 1.0f, 0.0f, 0.0f );

    // And some more spheres around it
    float tilt = -PI;

    for( int i = 1; i < OBJNUM; i++ ) {

        float frac = i/64.0f;

        si.obj_pos[i] = rotate( frac, tilt, DIST_INNER );
        si.obj_rad[i] = RAD_PLANET * RAD_PLANET;
        si.obj_col[i] = h2rgb( i/(float)OBJNUM );

        tilt += (PI*2.0f)/(float)(OBJNUM-1);

    }

    bmp_argb *dstbmp = bmp_alloc( width, height );
    if( dstbmp == NULL ) return 1;

    // Run directly if 0 threads
    if( threads == 0 ) {
        uint64_t t0 = get_usec();

        trace_lines( &si, dstbmp, 0, height-1 );

        uint64_t t1 = get_usec();

        printf( "Time: %f ms\n", (t1-t0)/1000.0f );

        bmp_save( dstbmp, "out.bmp", true );

        bmp_free( dstbmp );

        return 0;
    }

    // Threads >= 1

    int rc = pthread_barrier_init( &bar, NULL, threads + 1 );
    if( rc != 0 ) {
        perror( "pthread_barrier_init" );
        return 1;
    }

    ThreadInfo *ti = malloc( sizeof(ThreadInfo)*threads );
    if( ti == NULL ) {
        perror( "malloc" );
        return 1;
    }

    for( int i = 0; i < threads; i++ ) {
        ti[i].startline = (i*height)/threads;
        ti[i].endline   = ((i+1)*height)/threads - 1;
        ti[i].dstbmp    = dstbmp;
        ti[i].si        = &si;
        ti[i].id        = i;

        int rc = pthread_create( &ti[i].tid, 0, thread_func, &ti[i] );
        if( rc != 0 ) {
            perror( "pthread_create" );
            return 1;
        }

    }

    pthread_barrier_wait( &bar );

    uint64_t t0 = get_usec();

    pthread_barrier_wait( &bar );

    // Threads are now running

    // Wait for them to complete
    for( int i = 0; i < threads; i++ ) {
        pthread_join( ti[i].tid, 0 );
    }

    uint64_t t1 = get_usec();

    printf( "Time: %f ms\n", (t1-t0)/1000.0f );

    bmp_save( dstbmp, "out.bmp", true );

    bmp_free( dstbmp );

    return 0;
}
